from typing import Optional, Protocol, Self

from ..with_base.types import T_in, T_out
from .pipelineable_facade import PipelineableFacade
from ..with_base.stage_callable import StageCallableType
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..with_pipeline_processor.processor_interface import ProcessorInterface


class PipelineInterfaceMixin(
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    def build(
        self,
        processor: Optional[ProcessorInterface[T_in, T_out]] = None,
    ) -> PipelineableFacade[T_in, T_out]: ...

    def add(self, item: StageCallableType) -> Self: ...



from typing import Protocol

from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..partials.with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..partials.with_builder_methods.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithBuilderMethodsInterface,
)


class PipelineBuilderInterface(
    WithBuilderMethodsInterface[T_in, T_out],
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    pass



from abc import ABC
from typing import Any, Generic, Optional, Self

from ..with_base.types import T_in, T_out
from .pipelineable_facade import TPipelineable
from ..with_base.stage_callable import StageCallableType
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_pipeline_processor.processor_interface import ProcessorInterface


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    Generic[TPipelineable, T_in, T_out],
    ABC,
):
    pipelineable_class: type[TPipelineable]

    def __init__(
        self,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)

        self.set_items(self.get_items() or [])

    def add(self, item: StageCallableType) -> Self:
        return self.add_item(item)

    def build(
        self,
        processor: Optional[ProcessorInterface[T_in, T_out]] = None,
    ) -> TPipelineable:
        pipelineable = self.pipelineable_class()

        return pipelineable.set_items(
            [*self.get_items()],
        ).set_processor(
            processor or pipelineable.get_processor(),
        )



from typing import Any, cast

from ..partials.with_base.stage_callable import StageCallableType
from ..partials.with_base.types import T_in, T_out
from ..partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..partials.with_pipeline_processor.processor import (
    Processor as WithProcessorConcern,
)


class ChainedProcessor(
    WithProcessorConcern[T_in, T_out],
    ImplementsProcessorInterface[T_in, T_out],
):
    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        for stage in stages:
            payload = await self._call_stage(
                payload=payload, stage=stage, *args, **kwds
            )

        return cast(T_out, payload)



import pytest

from thecodecrate_pipeline import (
    PipelineBuilder,
)
from .stubs.stub_stages_int import (
    AddOneStage,
    TimesThreeStage,
    TimesTwoStage,
)


@pytest.mark.asyncio
async def test_pipeline_builder():
    pipeline_builder = (
        (PipelineBuilder[int]())
        .add(AddOneStage())
        .add(TimesTwoStage())
        .add(lambda payload: payload + 1)
        .add(AddOneStage())
    )

    pipeline = pipeline_builder.build()

    assert await pipeline.process(1) == 6


@pytest.mark.asyncio
async def test_immutability():
    pipeline_builder = (
        (PipelineBuilder[int]())
        .add(AddOneStage())
        .add(TimesTwoStage())
        .add(AddOneStage())
    )

    pipeline = pipeline_builder.build()

    assert await pipeline.process(1) == 5

    new_builder = pipeline_builder.add(TimesThreeStage())
    new_pipeline = new_builder.build()

    assert await new_pipeline.process(1) == 15
    assert await pipeline.process(1) == 5



from .classes.pipeline_builder import PipelineBuilder
from .classes.pipeline_builder_interface import PipelineBuilderInterface
from .classes.pipeline import Pipeline
from .classes.pipeline_interface import PipelineInterface
from .classes.processor import Processor
from .classes.processor_interface import ProcessorInterface
from .classes.stage import Stage
from .classes.stage_interface import StageInterface
from .processors.chained_processor import ChainedProcessor
from .processors.interruptible_processor import InterruptibleProcessor
from .partials.with_base.types import T_in, T_out
from .partials.with_base.stage_callable import StageCallable
from .partials.with_base.pipeline import Pipeline as WithPipelineBase
from .partials.with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from .partials.with_base.stage import Stage as WithStageBase
from .partials.with_base.stage_interface import (
    StageInterface as WithStageBaseInterface,
)
from .partials.with_builder_methods.pipeline_mixin import (
    PipelineMixin as WithBuilderMethods,
)
from .partials.with_builder_methods.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithBuilderMethodsInterface,
)
from .partials.with_pipeline_as_list.pipeline_mixin import (
    PipelineMixin as WithPipelineAsList,
)
from .partials.with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from .partials.with_pipeline_processor.pipeline_mixin import (
    PipelineMixin as WithPipelineProcessor,
)
from .partials.with_pipeline_processor.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineProcessorInterface,
)
from .partials.with_pipeline_processor.processor import (
    Processor as WithProcessorBase,
)
from .partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)
from .partials.with_pipeline_as_stage.pipeline_mixin import (
    PipelineMixin as WithPipelineAsStage,
)
from .partials.with_pipeline_as_stage.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsStageInterface,
)
from .partials.with_pipeline_as_immutable.pipeline_mixin import (
    PipelineMixin as WithPipelineAsImmutable,
)
from .partials.with_pipeline_as_immutable.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsImmutableInterface,
)
from .partials.with_pipelineable_methods.pipeline_mixin import (
    PipelineMixin as WithPipelineableMethods,
)
from .partials.with_pipelineable_methods.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineableMethodsInterface,
)
from .partials.with_stage_as_callable.stage_mixin import (
    StageMixin as WithStageAsCallable,
)
from .partials.with_stage_as_callable.stage_interface_mixin import (
    StageInterfaceMixin as WithStageAsCallableInterface,
)


# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.12.0"

# Expose the public API
__all__ = [
    # Core
    "Pipeline",
    "PipelineInterface",
    "Stage",
    "StageInterface",
    "Processor",
    "ProcessorInterface",
    "PipelineBuilder",
    "PipelineBuilderInterface",
    # Types
    "T_in",
    "T_out",
    "StageCallable",
    # Processors
    "ChainedProcessor",
    "InterruptibleProcessor",
    # Partials
    "WithPipelineBase",
    "WithPipelineBaseInterface",
    "WithStageBase",
    "WithStageBaseInterface",
    "WithBuilderMethods",
    "WithBuilderMethodsInterface",
    "WithPipelineAsList",
    "WithPipelineAsListInterface",
    "WithPipelineProcessor",
    "WithPipelineProcessorInterface",
    "WithProcessorBase",
    "WithProcessorBaseInterface",
    "WithPipelineAsStage",
    "WithPipelineAsStageInterface",
    "WithPipelineAsImmutable",
    "WithPipelineAsImmutableInterface",
    "WithPipelineableMethods",
    "WithPipelineableMethodsInterface",
    "WithStageAsCallable",
    "WithStageAsCallableInterface",
]



from abc import ABC
from typing import Self

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.stage_callable import StageCallableType
from ...support.act_as_list import ActAsList


class PipelineMixin(
    ActAsList[StageCallableType],
    ImplementsPipelineInterface,
    ABC,
):
    stage_instances: list[StageCallableType] = []

    def get_items(self) -> list[StageCallableType]:
        return self.stage_instances

    def set_items(self, items: list[StageCallableType]) -> Self:
        self.stage_instances = items

        return self



import inspect
from typing import Any, Awaitable, Callable, cast

from ..partials.with_base.stage_callable import StageCallableType
from ..partials.with_base.types import T_in, T_out
from ..partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..partials.with_pipeline_processor.processor import (
    Processor as WithProcessorConcern,
)

CheckCallable = Callable[[T_in], bool | Awaitable[bool]]


class InterruptibleProcessor(
    WithProcessorConcern[T_in, T_out],
    ImplementsProcessorInterface[T_in, T_out],
):
    check: CheckCallable[T_in]

    def __init__(self, check: CheckCallable[T_in]) -> None:
        self.check = check

    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        for stage in stages:
            payload = await self._call_stage(
                payload=payload, stage=stage, *args, **kwds
            )

            if not await self._call_check(payload):
                return cast(T_out, payload)

        return cast(T_out, payload)

    async def _call_check(self, payload: T_in) -> bool:
        result = self.check(payload)

        if inspect.isawaitable(result):
            return await result

        return result



from .pipeline import Pipeline
from .pipeline_builder_interface import (
    PipelineBuilderInterface as ImplementsPipelineBuilderInterface,
)
from ..partials.with_base.types import T_in, T_out
from ..partials.with_base.pipeline import (
    Pipeline as WithPipelineBaseConcern,
)
from ..partials.with_pipeline_as_list.pipeline_mixin import (
    PipelineMixin as WithPipelineAsListConcern,
)
from ..partials.with_builder_methods.pipeline_mixin import (
    PipelineMixin as WithBuilderMethodsConcern,
)


class PipelineBuilder(
    WithBuilderMethodsConcern[Pipeline[T_in, T_out], T_in, T_out],
    WithPipelineAsListConcern,
    WithPipelineBaseConcern,
    ImplementsPipelineBuilderInterface[T_in, T_out],
):
    pipelineable_class = Pipeline



from abc import ABC
from typing import Optional, Self

from .processor_interface import ProcessorInterface
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as PipelineInterface,
)
from ..with_base.types import T_in, T_out


class PipelineMixin(
    PipelineInterface[T_in, T_out],
    ABC,
):
    processor_class: Optional[type[ProcessorInterface[T_in, T_out]]] = None
    processor: Optional[ProcessorInterface[T_in, T_out]] = None

    async def process(
        self,
        payload: T_in,
    ) -> T_out:
        return await self.get_processor().process(
            payload=payload,
            stages=self.get_items(),
        )

    def get_processor(self) -> ProcessorInterface[T_in, T_out]:
        if self.processor:
            return self.processor

        if self.processor_class is None:
            raise ValueError("No processor set")

        if self.processor is None:
            self.processor = self.processor_class()

        return self.processor

    def set_processor(
        self, processor: ProcessorInterface[T_in, T_out]
    ) -> Self:
        self.processor = processor

        return self



from abc import ABC
from typing import Any, Self, Optional

from ..with_base.types import T_in, T_out
from ..with_pipeline_processor.processor_interface import ProcessorInterface
from ..with_base.stage_callable import StageCallableType
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    ABC,
):
    def __init__(
        self,
        processor: Optional[ProcessorInterface[T_in, T_out]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)

        self.set_items(self.get_items() or []).set_processor(
            processor or self.get_processor()
        )

    def pipe(self, stage: StageCallableType) -> Self:
        return self.add_item(stage)



from abc import abstractmethod
from typing import Any, Protocol, Self

from .processor_interface import ProcessorInterface
from ..with_base.types import T_in, T_out
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)


class PipelineInterfaceMixin(
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    @abstractmethod
    async def process(
        self,
        payload: T_in,
        *args: Any,
        **kwds: Any,
    ) -> T_out: ...

    def get_processor(
        self,
    ) -> ProcessorInterface[T_in, T_out]: ...

    def set_processor(
        self,
        processor: ProcessorInterface[T_in, T_out],
    ) -> Self: ...
